"""Type stubs for croniters Rust extension."""

MINUTE_FIELD: int
HOUR_FIELD: int
DAY_FIELD: int
MONTH_FIELD: int
DOW_FIELD: int
SECOND_FIELD: int
YEAR_FIELD: int

M_ALPHAS: dict[str, int]
DOW_ALPHAS: dict[str, int]
UNIX_FIELDS: list[int]
SECOND_FIELDS: list[int]
YEAR_FIELDS: list[int]
